/*
**   IRC - Internet Relay Chat, source/hash.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1991 Darren Reed
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "hash.h"

RCSID("$Id: hash.c,v 1.7 1999/02/21 20:28:42 cbehrens Exp $");

static IRCD_CLIENT		*CLIENT_HASH[HASH_CLIENT_SIZE+1];
static IRCD_CHANNEL		*CHANNEL_HASH[HASH_CHANNEL_SIZE];
static unsigned int		*_hashtab = NULL;

static unsigned int hash_compute_client(char *nname, unsigned int *hashv);
static unsigned int hash_compute_channel(char *cname, unsigned int *hashv);
static void add_client_to_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client);
static void del_client_from_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client);
static void add_channel_to_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);
static void del_channel_from_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);
static void add_channel_to_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);
static void del_channel_from_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);

void hash_init(void)
{
	int i;

	_hashtab = (u_int *) MyMalloc(256 * sizeof(unsigned int));
	for (i = 0; i < 256; i++)
		_hashtab[i] = tolower((char)i) * 109;
	BZERO(CLIENT_HASH, sizeof(CLIENT_HASH));
	BZERO(CHANNEL_HASH, sizeof(CHANNEL_HASH));
}

static unsigned int hash_compute_client(char *nname, unsigned int *hashv)
{
	unsigned char	*name = (unsigned char *)nname;
	unsigned int	hash = 1;

	for(;*name;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name];
	}
	*hashv = hash;
	if (*nname == '*')
		return HASH_CLIENT_SIZE;
	return hash % HASH_CLIENT_SIZE;
}

unsigned int hash_compute_whowas(char *nname)
{
	unsigned char	*name = (unsigned char *)nname;
	unsigned int	hash = 1;

	for(;*name;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name];
	}
	return hash;
}

static unsigned int hash_compute_channel(char *cname, unsigned int *hashv)
{
	unsigned char	*name = (unsigned char *)cname;
	unsigned int	hash = 1;
	int				i = 30;

	for(;*name && --i;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name] + (i << 1);
	}
	*hashv = hash;
	return hash % HASH_CHANNEL_SIZE;
}


IRCD_CLIENT *hash_find_client(char *name)
{
	unsigned int	hashv;
	unsigned int	hv;
	IRCD_CLIENT		*cptr;

	hv = hash_compute_client(name, &hashv);
	for(cptr=CLIENT_HASH[hv];cptr;cptr=cptr->h_next)
		if ((cptr->hashv == hashv) && !mycmp(name, cptr->name))
			break;
	return cptr;
}

IRCD_CLIENT *hash_find_chasing(char *name)
{
	IRCD_CLIENT	*cptr;

	cptr = hash_find_client(name);
	if (!cptr)
		cptr = whowas_get_history(name, 60);
	return cptr;
}

IRCD_CLIENT *hash_find_client_incl_hm(char *name)
{
	IRCD_CLIENT		*cptr;
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_client(name, &hashv);
	for(cptr=CLIENT_HASH[hv];cptr;cptr=cptr->h_next)
		if ((cptr->hashv == hashv) && !mycmp(name, cptr->name))
			break;
	if (!cptr && (hv != HASH_CLIENT_SIZE))
	{
		for(cptr=CLIENT_HASH[HASH_CLIENT_SIZE];cptr;cptr=cptr->h_next)
			if (!match(cptr->name, name))
				break;
	}
	return cptr;
}

IRCD_CHANNEL *hash_find_channel(char *name)
{
	IRCD_CHANNEL	*chptr;
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_channel(name, &hashv);
	for(chptr=CHANNEL_HASH[hv];chptr;chptr=chptr->h_next)
		if ((chptr->hashv == hashv) && !mycmp(name, chptr->name))
			break;
	return chptr;
}

int hash_add_client(IRCD_CLIENT *cptr)
{
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_client(cptr->name, &hashv);
	if (hv == HASH_CLIENT_SIZE)
		cptr->flags |= CLIENT_FLAGS_HOSTMASK;
	cptr->hashv = hashv;
	add_client_to_hlist(&CLIENT_HASH[hv], cptr);
	return 0;
}

int hash_del_client(IRCD_CLIENT *cptr)
{
	del_client_from_hlist(&CLIENT_HASH[ (cptr->flags & CLIENT_FLAGS_HOSTMASK) ?
				HASH_CLIENT_SIZE : cptr->hashv % HASH_CLIENT_SIZE ], cptr);
	return 0;
}

int hash_add_channel(IRCD_CHANNEL *chptr)
{
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_channel(chptr->name, &hashv);
	chptr->hashv = hashv;
	add_channel_to_hlist(&CHANNEL_HASH[hv], chptr);
	add_channel_to_list(&Channels, chptr);
	return 0;
}

int hash_del_channel(IRCD_CHANNEL *chptr)
{
	del_channel_from_hlist(&CHANNEL_HASH[chptr->hashv % HASH_CHANNEL_SIZE],
			chptr);
	del_channel_from_list(&Channels, chptr);
	return 0;
}

static void add_channel_to_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	channel->h_prev = NULL;
	if ((channel->h_next = *bucket) != NULL)
		channel->h_next->h_prev = channel;
	*bucket = channel;
}

static void del_channel_from_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	if (channel->h_prev)
		channel->h_prev->h_next = channel->h_next;
	else
		*bucket = channel->h_next;
	if (channel->h_next)
		channel->h_next->h_prev = channel->h_prev;
}

static void add_client_to_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->h_prev = NULL;
	if ((client->h_next = *bucket) != NULL)
		client->h_next->h_prev = client;
	*bucket = client;
}

static void del_client_from_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->h_prev)
		client->h_prev->h_next = client->h_next;
	else
		*bucket = client->h_next;
	if (client->h_next)
		client->h_next->h_prev = client->h_prev;
}

void add_client_to_llist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->l_prev = NULL;
	if ((client->l_next = *bucket) != NULL)
		client->l_next->l_prev = client;
	*bucket = client;
}

void del_client_from_llist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->l_prev)
		client->l_prev->l_next = client->l_next;
	else
		*bucket = client->l_next;
	if (client->l_next)
		client->l_next->l_prev = client->l_prev;
}

void add_client_to_lolist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->lo_prev = NULL;
	if ((client->lo_next = *bucket) != NULL)
		client->lo_next->lo_prev = client;
	*bucket = client;
}

void del_client_from_lolist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->lo_prev)
		client->lo_prev->lo_next = client->lo_next;
	else
		*bucket = client->lo_next;
	if (client->lo_next)
		client->lo_next->lo_prev = client->lo_prev;
}

void add_client_to_cllist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->cl_prev = NULL;
	if ((client->cl_next = *bucket) != NULL)
		client->cl_next->cl_prev = client;
	*bucket = client;
}

void del_client_from_cllist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->cl_prev)
		client->cl_prev->cl_next = client->cl_next;
	else
		*bucket = client->cl_next;
	if (client->cl_next)
		client->cl_next->cl_prev = client->cl_prev;
}

void add_client_to_ilist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->i_prev = NULL;
	if ((client->i_next = *bucket) != NULL)
		client->i_next->i_prev = client;
	*bucket = client;
}

void del_client_from_ilist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->i_prev)
		client->i_prev->i_next = client->i_next;
	else
		*bucket = client->i_next;
	if (client->i_next)
		client->i_next->i_prev = client->i_prev;
}

void add_client_to_oslist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->oslist = bucket;
	client->os_prev = NULL;
	if ((client->os_next = *bucket) != NULL)
		client->os_next->os_prev = client;
	*bucket = client;
}

void add_client_to_nlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->n_prev = NULL;
	if ((client->n_next = *bucket) != NULL)
		client->n_next->n_prev = client;
	*bucket = client;
}

void del_client_from_nlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->n_prev)
		client->n_prev->n_next = client->n_next;
	else
		*bucket = client->n_next;
	if (client->n_next)
		client->n_next->n_prev = client->n_prev;
}

void del_client_from_oslist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->oslist = NULL;
	if (client->os_prev)
		client->os_prev->os_next = client->os_next;
	else
		*bucket = client->os_next;
	if (client->os_next)
		client->os_next->os_prev = client->os_prev;
}

void add_client_to_ulist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->u_prev = NULL;
	if ((client->u_next = *bucket) != NULL)
		client->u_next->u_prev = client;
	*bucket = client;
}

void del_client_from_ulist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->u_prev)
		client->u_prev->u_next = client->u_next;
	else
		*bucket = client->u_next;
	if (client->u_next)
		client->u_next->u_prev = client->u_prev;
}

#if 0
void add_client_to_slist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->s_prev = NULL;
	if ((client->s_next = *bucket) != NULL)
		client->s_next->s_prev = client;
	*bucket = client;
}

void del_client_from_slist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->s_prev)
		client->s_prev->s_next = client->s_next;
	else
		*bucket = client->s_next;
	if (client->s_next)
		client->s_next->s_prev = client->s_prev;
}
#endif

void add_client_to_flist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->f_prev = NULL;
	if ((client->f_next = *bucket) != NULL)
		client->f_next->f_prev = client;
	*bucket = client;
}

void del_client_from_flist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->f_prev)
		client->f_prev->f_next = client->f_next;
	else
		*bucket = client->f_next;
	if (client->f_next)
		client->f_next->f_prev = client->f_prev;
}

void add_clientlist_to_list(IRCD_CLIENTLIST **bucket, IRCD_CLIENTLIST *clientlist)
{
	clientlist->prev = NULL;
	if ((clientlist->next = *bucket) != NULL)
		clientlist->next->prev = clientlist;
	*bucket = clientlist;
}

void del_clientlist_from_list(IRCD_CLIENTLIST **bucket, IRCD_CLIENTLIST *clientlist)
{
	if (clientlist->prev)
		clientlist->prev->next = clientlist->next;
	else
		*bucket = clientlist->next;
	if (clientlist->next)
		clientlist->next->prev = clientlist->prev;
}

void add_channellist_to_list(IRCD_CHANNELLIST **bucket, IRCD_CHANNELLIST *channellist)
{
	channellist->prev = NULL;
	if ((channellist->next = *bucket) != NULL)
		channellist->next->prev = channellist;
	*bucket = channellist;
}

void del_channellist_from_list(IRCD_CHANNELLIST **bucket, IRCD_CHANNELLIST *channellist)
{
	if (channellist->prev)
		channellist->prev->next = channellist->next;
	else
		*bucket = channellist->next;
	if (channellist->next)
		channellist->next->prev = channellist->prev;
}

void add_invite_to_list(IRCD_INVITE **bucket, IRCD_INVITE *invite)
{
	invite->prev = NULL;
	if ((invite->next = *bucket) != NULL)
		invite->next->prev = invite;
	*bucket = invite;
}

void del_invite_from_list(IRCD_INVITE **bucket, IRCD_INVITE *invite)
{
	if (invite->prev)
		invite->prev->next = invite->next;
	else
		*bucket = invite->next;
	if (invite->next)
		invite->next->prev = invite->prev;
}

void add_ban_to_list(IRCD_BAN **bucket, IRCD_BAN *ban)
{
	ban->prev = NULL;
	if ((ban->next = *bucket) != NULL)
		ban->next->prev = ban;
	*bucket = ban;
}

void del_ban_from_list(IRCD_BAN **bucket, IRCD_BAN *ban)
{
	if (ban->prev)
		ban->prev->next = ban->next;
	else
		*bucket = ban->next;
	if (ban->next)
		ban->next->prev = ban->prev;
}

static void add_channel_to_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	channel->prev = NULL;
	if ((channel->next = *bucket) != NULL)
		channel->next->prev = channel;
	*bucket = channel;
}

static void del_channel_from_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	if (channel->prev)
		channel->prev->next = channel->next;
	else
		*bucket = channel->next;
	if (channel->next)
		channel->next->prev = channel->prev;
}

