/*
**   IRC - Internet Relay Chat, source/ircd.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"

RCSID("$Id: ircd.c,v 1.34 1999/02/23 07:57:20 cbehrens Exp $");

int				Now;
char			*Conffile;
int				ircd_pagesize;
FDLIST			Fdlist_main;
static FDLIST	Fdlist_clients[MAXFDLISTS];
static int		fdlist_lasttime[MAXFDLISTS];
IRCD_CLIENT		Me, *meptr = &Me;
IRCD_CLIENT		*Fds[MAXCONNECTIONS];	/* for easy fd -> client lookup */
IRCD_CLIENT		*Locals = NULL;			/* local clients */
IRCD_CLIENT		*Clients = NULL;
IRCD_CLIENT		*Opers = NULL;
IRCD_CLIENT		*Servers = NULL;
IRCD_CLIENT		*Flooders = NULL;
IRCD_CHANNEL	*Channels = NULL;
GLOBALSTATS		Globalinfo;
IRCD_CLIENT		*Rejectholds = NULL;
IRCD_MOTD		*Motd = NULL;
int				DoRehash = 0;
char			*RehashWho = NULL;
int				Rehashed = 0;

static IRCD_CLIENT	_ircd_resclient;
static int			_ircd_resfd = -1;

static RETSIGTYPE sighup_handler(int sig);
static void res_read(IRCD_CLIENT **cptr, SERVLINK **slptr);
static int _ircd_pollfds(FDLIST *fdlist, int secs);
static void flood_check(void);
static void ping_check(void);
static void kline_check(void);
static void rejecthold_check(void);

static RETSIGTYPE sighup_handler(int sig)
{
	DoRehash = 1;
	if (!RehashWho)
		RehashWho = MyStrdup("SIGHUP");
	signal(sig, sighup_handler);
}

#ifndef MAP_ANON
int			Ircd_zero_fd = -1;
#endif

static void res_read(IRCD_CLIENT **clptr, SERVLINK **slptr)
{
	MYRES_STATUS	*sptr;
	MYRES_HOSTENT	*heptr;
	IRCD_CLIENT		*cptr = NULL;
	SERVLINK		*slink = NULL;

	*clptr = NULL;
	*slptr = NULL;
	sptr = myres_get_answer();
	if (!sptr)
		return;
	if (sptr->ctype == MYRES_TYPE_CLIENT)
	{
		cptr = (IRCD_CLIENT *)sptr->cptr;
		cptr->res_status.rptr = NULL;
	}
	else if (sptr->ctype == MYRES_TYPE_SERVLINK)
	{
		slink = (SERVLINK *)sptr->cptr;
		slink->res_status.rptr = NULL;
	}
	else
		return;
	if (!sptr->result) /* FAILED */
	{
		switch(sptr->ctype)
		{
			case MYRES_TYPE_CLIENT:
				cptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
				*clptr = cptr;
				break;
			case MYRES_TYPE_SERVLINK:
				send_to_opers("Lookup of %s for \"link %s\" conf entry failed",
					slink->hostname, slink->servername);
				slink->flags &= ~SERVLINK_FLAGS_DOINGDNS;
				*slptr = slink;
				break;
		}
		return;
	}
	if (sptr->ctype == MYRES_TYPE_CLIENT)
	{
		if (sptr->type == MYRES_TYPE_PTR)
		{
			myres_hostent_move(&(cptr->he_reverse), &(sptr->he));
			sptr->type = 0;
			heptr = myres_gethostbyname(cptr->he_reverse.h.h_name, sptr);
			if (heptr)
			{
				myres_hostent_move(&(cptr->he_forward), &(sptr->he));
				MyFree(heptr);
			}
			if (!heptr)
			{
				*clptr = cptr;
				return;
			}
		}
		else
			myres_hostent_move(&(cptr->he_forward), &(sptr->he));
		myres_status_deinit(&(cptr->res_status));

		/* check reverse vs forward */
		if (sptr->result)
		{
			ircd_uint32_t	ipaddr = htonl(cptr->rem_ipaddr);
			char			**addrs = cptr->he_forward.h.h_addr_list;

			for(;*addrs;addrs++)
				if (!BCMP(*addrs, (char *)&ipaddr, sizeof(struct in_addr)))
					break;
			if (!*addrs)
			{
				DEBUG(DEBUG_ALL, ("IP Mismatch: %s != %s[%s]",
					cptr->hostname, cptr->he_forward.h.h_name,
					cptr->he_forward.h.h_addr_list[0] ? 
					inet_ntoa(*(struct in_addr *)
								cptr->he_forward.h.h_addr_list[0]) :
					"null"));
			}
			else
			{
				if (strlen(cptr->he_forward.h.h_name) <=
								sizeof(cptr->hostname)-1)
					strcpy(cptr->hostname, cptr->he_forward.h.h_name);
			}
		}
		cptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
		*clptr = cptr;
		return;
	}
	/* MYRES_TYPE_SERVLINK */
	if (sptr->type != MYRES_TYPE_A)
	{
		myres_status_deinit(&(slink->res_status));
		slink->flags &= ~SERVLINK_FLAGS_DOINGDNS;
		*slptr = slink;
		return;
	}
	myres_hostent_move(&(slink->he_forward), &(sptr->he));
	myres_status_deinit(&(slink->res_status));
	slink->flags &= ~SERVLINK_FLAGS_DOINGDNS;
	if (slink->ipaddr == -1)
	{
        if (slink->he_forward.h.h_addr_list[0])
        {
            BCOPY(slink->he_forward.h.h_addr_list[0], &slink->ipaddr, sizeof(slink->ipaddr));
            slink->ipaddr = htonl(slink->ipaddr);
        }
	}
	*slptr = slink;
	return;
}

static int _ircd_pollfds(FDLIST *fdlist, int secs)
{
#ifdef HAVE_POLL
	static FDLIST_POLLFD pfdarray[MAXCONNECTIONS];
#else
	fd_set			readfds;
	fd_set			writefds;
	struct timeval	tv;
#endif
	int				lowestfd;
	int				i;
	int				numfds;
	int				ret;
	static char		readbuf[1024*64];
	static int		readbufsize = 1024*64;
	IRCD_CLIENT		*cptr;
	char			exitbuf[512];

	if (!fdlist->pfd_num)
		return 0;

#ifdef HAVE_POLL
	BCOPY(fdlist->pfd_array, pfdarray, sizeof(FDLIST_POLLFD)*fdlist->pfd_num); 
	numfds = fdlist->pfd_num;
	lowestfd = 0;
#else
	bcopy(&fdlist->read_fds, &readfds, sizeof(readfds));
	bcopy(&fdlist->write_fds, &writefds, sizeof(writefds));
	numfds = MAXCONNECTIONS;
	lowestfd = fdlist->lowest_fd;
#endif
	for(;;) /* fake loop */
	{
#ifdef HAVE_POLL
		ret = poll(pfdarray, numfds, secs*1000);
#else
		tv.tv_sec = secs;
		tv.tv_usec = 0;
		ret = select(numfds+1, &readfds, &writefds, NULL, &tv);
#endif
		if (ret < 0)
		{
			Now = time(NULL);
			return ret;
		}
		Now = time(NULL);
		for(i=lowestfd;ret && i<numfds;i++)
		{
#ifdef HAVE_POLL
			if (!pfdarray[i].revents)
#else
			if (!FD_ISSET(i, &readfds) && !FD_ISSET(i, &writefds))
#endif
				continue;
			ret--;
#ifdef HAVE_POLL
			if ((cptr = GET_FDS(pfdarray[i].fd)) == NULL)
#else
			if ((cptr = GET_FDS(i)) == NULL)
#endif
				continue;
			cptr->lasttime = Now;
			if (IsRes(cptr))
			{
				SERVLINK	*slink;

				res_read(&cptr, &slink);
				if (cptr)
				{
					if (!(cptr->flags &
								(CLIENT_FLAGS_DOINGAUTH|CLIENT_FLAGS_DOINGDNS)))
					{
						if (ircd_parse(cptr) < 0)
						{
							continue;
						}
					}
				}
				else if (slink)
				{

				}
				continue;
			}
#ifdef HAVE_POLL
			if (pfdarray[i].revents & (POLLERR|POLLHUP|POLLNVAL))
				pfdarray[i].revents |= POLLIN; /* will close() upon read()*/
			if (pfdarray[i].revents & POLLIN)
#else
			if (FD_ISSET(i, &readfds))
#endif
			{
				if (IsListener(cptr))
				{
					struct sockaddr_in	sin;
					int					sinsize = sizeof(sin);

					ret = accept(cptr->fdlistinfo.fd,
								(struct sockaddr *)&sin, &sinsize);
					if (ret < 0)
						continue;
					if (ret >= MAXCONNECTIONS)
						close(ret);
					else
					{
						int ipaddr = ntohl(sin.sin_addr.s_addr);

						if (iphash_handle_accept(ipaddr))
						{
							int num;

							if ((num=(int)vars_value_get_bynum(IRCD_VAR_REJECT_HOLD)) > 0)
							{
								if (client_add_rejecthold(ret, &sin) >= 0)
									continue;
							}
							close(ret);
							continue;
						}

						if (!access_match_ipaddr(&(_ircd_conf.deny),
							NULL, NULL, ipaddr, 0, NULL))
						{
							DEBUG(DEBUG_ALL, ("denying accept from %s\n",
								inet_ntoa(sin.sin_addr)));
							close(ret);
						}
						else if (client_add_local(cptr, ret, &sin) < 0)
							close(ret);
					}
					continue;
				}
				ret = read(cptr->fdlistinfo.fd, readbuf, readbufsize-1);
				if ((ret < 0) && 
						((errno == EAGAIN) || (errno == EINTR)
							|| (errno == EWOULDBLOCK)))
					continue;
				if (ret <= 0)
				{
					IRCD_CLIENT *cptrkeep = cptr->auth_ptr;

					if (ret == 0)
						sprintf(exitbuf, "EOF of client");
					else
						sprintf(exitbuf, "Read error %i: %s", errno,
							(errno == 0) ? "Undefined error" : 
									support_strerror(errno));
					if (!IsAuth(cptr))
					{
						client_del(meptr, cptr, exitbuf);
						continue;
					}
					client_del(meptr, cptr, exitbuf);
					cptr = cptrkeep; /* fall through to the ircd_parse */
				}
				else
					readbuf[ret] = '\0';
				if (IsRejecthold(cptr))
					continue;
				if (ret > 0 && (queue_data_add(&(cptr->queue_in), readbuf, ret) < 0))
				{
					IRCD_CLIENT *cptrkeep = cptr->auth_ptr;

					if (!IsAuth(cptr))
					{
						client_del(meptr, cptr, "Couldn't add to RecvQ");
						continue;
					}
					client_del(meptr, cptr, "Couldn't add to RecvQ");
					cptr = cptrkeep; /* fall through to the ircd_parse */
				}
				else if (ret > 0)
				{
					cptr->recv_bytes += ret;
					if (cptr->recv_bytes >= 1024)
					{
						cptr->recv_kbytes += cptr->recv_bytes / 1024;
						cptr->recv_bytes %= 1024;
					}
					if (IsClient(cptr))
						Globalinfo.bw_c_incoming_bytes += ret;
					else if (IsServer(cptr))
						Globalinfo.bw_s_incoming_bytes += ret;
				}
				if (IsAuth(cptr))
				{
					char	*buf = NULL;
					int		len = 0;
					short	remp, locp;
					char	ruser[30];
					IRCD_CLIENT	*cptrkeep;
					char	tuser[128];
					char	*blah = tuser;

					if (queue_data_get(&(cptr->queue_in), &buf, &len) < 0)
						continue;

					if (len <= sizeof(tuser)-1)
					{
						strncpy(blah, buf, len);
					}
					else
					{
						blah = (char *)MyMalloc(len+1);
						strncpy(blah, buf, len);
					}
					blah[len] = '\0';

					strncpy(tuser, buf, sizeof(tuser)-1);
					if (len < sizeof(tuser))
						tuser[len] = '\0';
					else
						tuser[sizeof(tuser)-1] = '\0';
					if ((len > 10) && sscanf(tuser, "%hd , %hd : USERID : %*[^:]: %10s",
						&remp, &locp, ruser) == 3)
					{
						char *s = strrchr(tuser, ':') + 1;
						char *t;

						len = 0;
						for(t=ruser;*s && (len<(sizeof(ruser)-1));s++)
							if (!isspace(*s) && (*s != ':') && (*s != '@'))
							{
								len++;
								*t++ = *s;
							}
						*t = '\0';
						strncpy(cptr->auth_ptr->authusername, ruser,
								sizeof(cptr->auth_ptr->authusername)-1);
						strncpy(cptr->auth_ptr->username, ruser,
								sizeof(cptr->auth_ptr->username)-1);
						if (*ruser)
							cptr->auth_ptr->flags |= CLIENT_FLAGS_GOTAUTH;
					}
					else
					{
						if (!strchr(buf, '\n') && !strchr(buf, '\r'))
						{
							if (blah != tuser)
								MyFree(blah);
							continue;
						}
					}
					cptrkeep = cptr->auth_ptr;
					client_del_local(cptr, cptr, NULL);
					cptr = cptrkeep; /* fall through to the ircd_parse */
					if (blah != tuser)
						MyFree(blah);
				}
				if (!IsServer(cptr) && (cptr->queue_in.queue_size > 8192))
				{
					IRCD_CLIENT *cptrkeep = cptr->auth_ptr;

					if (!IsAuth(cptr))
					{
						client_del(meptr, cptr, "Excess Flood");
						continue;
					}
					client_del(meptr, cptr, "Excess Flood");
					cptr = cptrkeep; /* fall through to the ircd_parse */
				}
				if (!(cptr->flags &
							(CLIENT_FLAGS_DOINGAUTH|CLIENT_FLAGS_DOINGDNS)))
				{
					if (ircd_parse(cptr) < 0)
					{
						continue;
					}
				}
			}
#ifdef HAVE_POLL
			if (pfdarray[i].revents & POLLOUT)
#else
			if (FD_ISSET(i, &writefds))
#endif
			{
				if (cptr->flags & CLIENT_FLAGS_DOINGCONNECT)
				{
					cptr->flags &= ~CLIENT_FLAGS_DOINGCONNECT;
					/* connected to wherever */
					if (IsAuth(cptr))
					{
						char buffer[64];
						int	len;

						sprintf(buffer, "%u , %u\r\n",
							(unsigned int)cptr->auth_ptr->rem_port,
							(unsigned int)cptr->auth_ptr->loc_port);
						len = strlen(buffer);
						if (write(cptr->fdlistinfo.fd, buffer, len) != len)
						{
							cptr->auth_ptr->flags &= ~CLIENT_FLAGS_DOINGAUTH;
							queue_delete(&(cptr->queue_in));
							ircd_close_fd(&(cptr->fdlistinfo));
							cptr->auth_ptr->auth_ptr = NULL;
							MyFree(cptr);
							continue;
						}
						fdlist_del_event(cptr->fdlistinfo.fdlist,
								cptr->fdlistinfo.fdlist_pos,
								FDLIST_FLAGS_WRITE);
						continue;
					}
					else
					{
						if (!cptr->queue_out.queue_size) /* should be false */
							fdlist_del_event(cptr->fdlistinfo.fdlist,
								cptr->fdlistinfo.fdlist_pos,
								FDLIST_FLAGS_WRITE);
					}
				}
				if (cptr->queue_out.queue_size > cptr->maxsendq)
				{
					sprintf(exitbuf, "Max SendQ Exceeded: %i > %i",
							cptr->queue_out.queue_size, cptr->maxsendq);
					client_del(meptr, cptr, exitbuf);
					continue;
				}
				if (client_queue_flush(cptr, exitbuf))
					client_del(meptr, cptr, exitbuf);
			}
		}
		break;
	}
	return 0;
}

static void flood_check(void)
{
	IRCD_CLIENT	*fptr;
	IRCD_CLIENT	*nextf;

	for(fptr=Flooders;fptr;fptr=nextf)
	{
		nextf = fptr->f_next;
		if (Now > fptr->rate_time)
			ircd_parse(fptr);
	}
}

static void ping_check(void)
{
	IRCD_CLIENT	*pptr;
	IRCD_CLIENT	*nextlo;

	for(pptr=Locals;pptr;pptr=nextlo)
	{
		nextlo = pptr->lo_next;

		if (IsDead(pptr))
		{
			client_del(meptr, pptr, "Dead client");
			continue;
		}

		if (IsDoingDns(pptr) && (Now <= pptr->firsttime + 30))
		{
			if (myres_retry_request(pptr->res_status.rptr))
			{
				myres_status_deinit(&(pptr->res_status));
				pptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
			}
			else
			{
				continue;
			}
		}
		if (IsUnknown(pptr) && (Now > pptr->firsttime + 30))
		{
			if (IsDoingConnect(pptr) || (!IsDoingDns(pptr) && !IsDoingAuth(pptr)))
			{
				client_del(meptr, pptr, "Too much time as unknown");
				continue;
			}
			if (IsDoingAuth(pptr))
			{
				if (pptr->auth_ptr) /* assume this should be true! */
					client_del(meptr, pptr->auth_ptr, "Auth timeout.");
			}
			if (IsDoingDns(pptr))
			{
				myres_status_deinit(&(pptr->res_status));
				pptr->flags &= ~CLIENT_FLAGS_DOINGDNS;
			}
			pptr->firsttime = Now;
			ircd_parse(pptr);
			continue;
		}
		if (pptr->pingfreq && ((Now - pptr->lasttime) > pptr->pingfreq))
		{
			if (pptr->flags & CLIENT_FLAGS_PINGSENT)
			{
				if ((Now - pptr->lasttime) > (pptr->pingfreq*2))
				{
					char buffer[128];

					sprintf(buffer,
						"Ping timeout: no data for %i seconds",
						Now - pptr->lasttime);
					
					client_del(meptr, pptr, buffer);
					continue;
				}
			}
			else
			{
				send_to_client(pptr, "PING :%s", meptr->name);
				pptr->flags |= CLIENT_FLAGS_PINGSENT;
			}
		}
		/* fdlist check */
		if (IsClient(pptr))
		{
			int		fdlistnum = Now - pptr->idletime;
			FDLIST	*fdlist = &Fdlist_main;
			int		mfd;
			int		ret;
			int		flags;
			int		nfdlists;
			int		ptime;

			if (!IsAnOper(pptr) && ((nfdlists = (int)vars_value_get_bynum(IRCD_VAR_PRIORITY_FDLISTS)) > 0))
			{
				ptime = (int)vars_value_get_bynum(IRCD_VAR_PRIORITY_IDLE);
				fdlistnum /= ptime;
				if (fdlistnum > 0)
				{
					fdlistnum--;
					if (fdlistnum >= nfdlists)
						fdlistnum = nfdlists - 1;
					fdlist = &Fdlist_clients[fdlistnum];
				}
			}
			if (pptr->fdlistinfo.fdlist != fdlist)
			{
				flags = fdlist_get_events(pptr->fdlistinfo.fdlist,
					pptr->fdlistinfo.fdlist_pos);
				assert(flags);
				ret = fdlist_del_fd(pptr->fdlistinfo.fdlist,
                        pptr->fdlistinfo.fdlist_pos, &mfd);
				if (mfd >= 0)
				{
					if (GET_FDS(mfd))
						GET_FDS(mfd)->fdlistinfo.fdlist_pos =
	                        pptr->fdlistinfo.fdlist_pos;
				}
				ret = ircd_open_fd(pptr->fdlistinfo.fd, flags, pptr, fdlist);
				if (ret < 0)
					client_del(meptr, pptr, "Fdlist problem");
			}
		}
	}
}

static void servlink_check(void)
{
	SERVLINK	*sptr;

	for(sptr=_ircd_conf.servers;sptr;sptr=sptr->next)
	{
		if ((sptr->flags & SERVLINK_FLAGS_DOINGDNS) &&
						myres_retry_request(sptr->res_status.rptr))
		{
			myres_status_deinit(&(sptr->res_status));
			sptr->flags &= ~SERVLINK_FLAGS_DOINGDNS;
			send_to_opers("Lookup of %s for \"link %s\" conf entry failed",
					sptr->hostname, sptr->servername);
		}
	}
}

static void kline_check(void)
{
	IRCD_CLIENT		*pptr;
	IRCD_CLIENT		*nextlo;
	int				ret;
	char			exitbuf[512];
	CONFHOSTIPENTRY	*entryptr;
	char			cnamebuf[CLIENTNAMESIZE];

	for(pptr=Locals;pptr;pptr=nextlo)
	{
		nextlo = pptr->lo_next;

		if (!IsClient(pptr))
			continue;

		if (pptr->class->flags & CLASS_FLAGS_BYPASSBANS)
			continue;

		ret = access_match_ipaddr(&(_ircd_conf.deny),
				pptr->username, pptr->ipstring,
				pptr->rem_ipaddr, ACCESS_FLAGS_BESTMATCH,
				&entryptr);
		if (ret < 0)
			ret = access_match_hostname(&(_ircd_conf.deny),
				pptr->username, pptr->hostname, ACCESS_FLAGS_BESTMATCH,
				&entryptr);

		if (ret >= 0) /* banned */
		{
			send_to_opers("Kill line active for %s",
					ircd_clientname(pptr, 0, cnamebuf));
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name),
					entryptr->data ? entryptr->data : "No reason specified");
			sprintf(exitbuf, "Banned: %s",
				entryptr->data ? entryptr->data : "No reason specified");
			client_del(meptr, pptr, exitbuf);
			continue;
		}
	}
}

static void rejecthold_check(void)
{
	IRCD_CLIENT	*rptr;
	IRCD_CLIENT	*nextrptr;
	int			num = (int)vars_value_get_bynum(IRCD_VAR_REJECT_HOLD);

	for(rptr=Rejectholds;rptr;rptr=rptr=nextrptr)
	{
		nextrptr = rptr->lo_next;
		if ((num <= 0) || ((rptr->firsttime+num) <= Now))
			client_del(meptr, rptr, "bye now");
	}
}

int main(int argc, char **argv)
{
	int				ret;
	register int	i;
	FDLIST			*curfdlist;
	int				nofork = 0;
#ifdef RLIMIT_NOFILE
	struct rlimit	r;
#endif

	if (argc < 2)
	{
		fprintf(stderr, "Usage: myircd ircd.conf\n");
		exit(-1);
	}
	
	if ((argc > 2) && (*argv[1] == '-'))
	{
		if (argv[1][1] == 'n')
		{
			nofork = 1;
			argc--;
			argv++;
		}
	}

#ifdef RLIMIT_NOFILE
	r.rlim_cur = MAXCONNECTIONS;
	r.rlim_max = MAXCONNECTIONS;
	setrlimit(RLIMIT_NOFILE, &r);
#endif

	{ /* I know this is lame, but I don't care. */
		int fds[MAXCONNECTIONS];

		for(i=0;i<MAXCONNECTIONS;i++)
		{
			ret = open("/dev/null", O_RDONLY);
			if (ret < 0)
				break;
			fds[i] = ret;
		}
		i--;
		if ((MAXCONNECTIONS < 1) || (i < 0) || (fds[i] < (MAXCONNECTIONS-1)))
		{
			int num = i < 0 ? 0 : fds[i]+1;

			fprintf(stderr,
			"MAXCONNECTIONS is set incorrectly.  I can only open %i file%s\n",
				num, num == 1 ? "" : "s");
			exit(-1);
		}
		while(i >= 0)
		{
			close(fds[i]);
			i--;
		}
	}

	openlog("ircd-comstud", LOG_PID|LOG_NDELAY, LOG_DAEMON);
#ifdef MALLOC_DEBUG
	dmalloc_init(32*1024-1,32*1024-1,8);
#endif
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, sighup_handler);
	Conffile = argv[1];
	for(i=0;i<MAXCONNECTIONS;i++)
		SET_FDS(i, NULL);
	_ircd_resfd = myres_init();
	if (_ircd_resfd < 0)
	{
		fprintf(stderr, "Couldn't do myres_init: %i\n", _ircd_resfd);
		exit(-1);
	}
	BZERO(&_ircd_resclient, sizeof(_ircd_resclient));
	_ircd_resclient.type = CLIENT_TYPE_RES;
#ifndef MAP_ANON
	Ircd_zero_fd = open("/dev/zero", O_RDWR);
	if (Ircd_zero_fd < 0)
	{
		fprintf(stderr, "Couldn't open /dev/zero\n");
		exit(-1);
	}
#endif
#if 0
	ircd_pagesize = ircd_getpagesize();
#else
	ircd_pagesize = 2*1024;
#endif
	vars_init();
	command_tree_init();
	numeric_init();
	ircdconf_init(&_ircd_conf);
	fdlist_init(&Fdlist_main);
	for(i=0;i<MAXFDLISTS;i++)
	{
		fdlist_init(&Fdlist_clients[i]);
		fdlist_lasttime[i] = 0;
	}
	hash_init();
	iphash_init();
	BZERO(Fds, sizeof(Fds));
	BZERO(&Globalinfo, sizeof(Globalinfo));
	Globalinfo.m_maxconntime = Globalinfo.g_maxclientstime = 
			Globalinfo.g_maxserverstime = Globalinfo.maxchannelstime =
			Globalinfo.m_maxclientstime = Globalinfo.m_maxserverstime =
				(int)time(NULL);
	Globalinfo.g_servers = 1;
	BZERO(&Me, sizeof(IRCD_CLIENT));
	if (ircd_open_fd(_ircd_resfd, FDLIST_FLAGS_READ, &_ircd_resclient, NULL) < 0)
	{
		fprintf(stderr, "Error putting resfd into fdlist\n");
		exit(-1);
	}
	Now = (int)time(NULL);
	ret = ircdconf_read(Conffile, &_ircd_conf);
	if (ret < 0)
	{
		fprintf(stderr, "Couldn't read config file \"%s\": %i\n",
					Conffile, ret);
		exit(ret);
	}
	if (!_ircd_conf.name)
	{
		fprintf(stderr, "No server name given in ircd.conf\n");
		exit(-1);
	}
	{
		char    *motd_file = vars_value_get_bynum(IRCD_VAR_MOTD_FILE);

		if (motd_file)
			motd_load(motd_file, &Motd);
	}

	whowas_init();
	strncpy(meptr->name, _ircd_conf.name, sizeof(meptr->name)-1);
	strncpy(meptr->ircname, _ircd_conf.info ? _ircd_conf.info : "Unknown",
			sizeof(meptr->ircname)-1);
	meptr->ircname[sizeof(meptr->ircname)-1] = '\0';
	meptr->type = CLIENT_TYPE_SERVER;
	meptr->from = meptr;
	meptr->servptr = meptr;
	meptr->firsttime = meptr->lasttime = Now;
	hash_add_client(meptr);
	add_client_to_oslist(&Servers, meptr);

	if (!_ircd_conf.listeners)
	{
		fprintf(stderr, "No ports on which to listen\n");
		exit(-1);
	}

	if (!nofork)
	{
		closelog();
		ret = fork();
		if (ret)
		{
			exit(0);
		}
		openlog("ircd-comstud", LOG_PID|LOG_NDELAY, LOG_DAEMON);
	}
	logs_init();
	syslog(LOG_NOTICE, "ircd has started.");

	{
		int fd = open("ircd.pid", O_CREAT|O_TRUNC|O_WRONLY, 0660);

		if (fd >= 0)
		{
			char buf[10];

			sprintf(buf, "%i\n", (int)getpid());
			write(fd, buf, strlen(buf));
			close(fd);
		}
	}
	curfdlist = &Fdlist_main;
	for(;;)
	{
		if (DoRehash)
		{
			ircdconf_rehash();
			DoRehash = 0;
		}

		ret = _ircd_pollfds(curfdlist, 2);

		if (Now >= (fdlist_lasttime[0] + (int)vars_value_get_bynum(IRCD_VAR_PRIORITY_LAG)))
		{
			_ircd_pollfds(&Fdlist_clients[0], 0);
			fdlist_lasttime[0] = Now;
			for(i=1;i<(int)vars_value_get_bynum(IRCD_VAR_PRIORITY_FDLISTS);i++)
			{
				if (Now >= (fdlist_lasttime[i] + ((i+1)*(int)vars_value_get_bynum(IRCD_VAR_PRIORITY_LAG))))
				{
					_ircd_pollfds(&Fdlist_clients[i], 0);
					fdlist_lasttime[i] = Now;
				}
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 2))
			{
				flood_check();
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 5))
			{
				ping_check();
				lastcheck = Now;
			}
		}

		if (Now >= (Globalinfo.bw_last_check + 5))
		{
			if (Globalinfo.bw_last_check)
			{
				Globalinfo.bw_c_incoming_rate = Globalinfo.bw_c_incoming_bytes/(Now - Globalinfo.bw_last_check);
				Globalinfo.bw_c_outgoing_rate = Globalinfo.bw_c_outgoing_bytes/(Now - Globalinfo.bw_last_check);
				Globalinfo.bw_s_incoming_rate = Globalinfo.bw_s_incoming_bytes/(Now - Globalinfo.bw_last_check);
				Globalinfo.bw_s_outgoing_rate = Globalinfo.bw_s_outgoing_bytes/(Now - Globalinfo.bw_last_check);
			}
			Globalinfo.bw_c_incoming_bytes = 0;
			Globalinfo.bw_c_outgoing_bytes = 0;
			Globalinfo.bw_s_incoming_bytes = 0;
			Globalinfo.bw_s_outgoing_bytes = 0;
			Globalinfo.bw_last_check = Now;
		}

		{
			static int	lastcheck = 0;

			if (Rehashed || (Now >= (lastcheck + 900)))
			{
				kline_check();
				lastcheck = Now;
				Rehashed = 0;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 5))
			{
				servlink_check();
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 10))
			{
				iphash_cleanup();
				lastcheck = Now;
			}
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 900))
			{
				queue_cleanup();
				lastcheck = Now;
			}	
		}

		{
			static int	lastcheck = 0;

			if (Now >= (lastcheck + 10))
			{
				rejecthold_check();
				lastcheck = Now;
			}
		}
	}
	/* NOT REACHED */
	exit(0);
}
