/*
**   IRC - Internet Relay Chat, source/res.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright 1992, 1993, 1994 Darren Reed.
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
**  This file was majorly cleaned up/partly rewritten by
**     Thomas Helvey <tomh@inxpress.net>
**  After that, I cleaned it up even further.  - CJB
**
*/

#include "ircd_includes.h"
#include <arpa/nameser.h>
#include <resolv.h>
#include <signal.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "res.h"

RCSID("$Id: res.c,v 1.8 1999/02/19 23:34:21 cbehrens Exp $");

/* dn_skipname is really an internal function,
   we shouldn't be using it in res.c */
#if !defined(dn_skipname) && !defined(__dn_skipname)
extern int dn_skipname(const u_char *, const u_char *);
#endif

#define MAXPACKET	1024

#define RES_MAXADDRS	35
#define RES_MAXALIASES	35

#define ADDRSBLEN ((RES_MAXADDRS * sizeof(char *)) + sizeof(char *))
#define ADDRSDLEN (RES_MAXADDRS * sizeof(struct in_addr))
#define ALIASBLEN ((RES_MAXALIASES * sizeof(char *)) + sizeof(char *))
#define MAXGETHOSTLEN (ADDRSBLEN + ADDRSDLEN + ALIASBLEN + MAXPACKET)

#define AR_TTL	600

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff                       
#endif                                                 
#ifndef INT16SZ        
#define INT16SZ 2         
#endif
#ifndef INT32SZ
#define INT32SZ 4
#endif            

#define MYRES_MAX(a, b)   ((a) > (b) ? (a) : (b))
#define MYRES_MIN(a, b)   ((a) < (b) ? (a) : (b))


/*
 * Building the Hostent
 * The Hostent struct is arranged like this:
 *          +-------------------------------+
 * Hostent: | struct hostent h              |
 *          |-------------------------------|
 *          | char* buf                     |
 *          +-------------------------------+
 *
 * allocated:
 *
 *          +-------------------------------+
 * buf:     | h_aliases pointer array   | 0 |
 *          |-------------------------------|
 *          | h_addr_list pointer array | 0 |
 *          |-------------------------------|
 *          | h_addr_list addresses         |
 *          |-------------------------------|
 *          | storage for hostname strings  |
 *          +-------------------------------+
 *
 *  For requests the size of the h_aliases, and h_addr_list pointer
 *  array sizes are set to MAXALISES and MAXADDRS respectively, and
 *  buf is a fixed size with enough space to hold the largest expected
 *  reply from a nameserver, see RFC 1034 and RFC 1035.
 *  For cached entries the sizes are dependent on the actual number
 *  of aliases and addresses. If new aliases and addresses are found
 *  for cached entries, the buffer is grown and the new entries are added.
 *  The hostent struct is filled in with the addresses of the entries in
 *  the Hostent buf as follows:
 *  h_name      - contains a pointer into the hostname string area.
 *  h_aliases   - contains a pointer to the start of the buffer.
 *  h_addr_list - contains a pointer to the end of h_aliases + sizeof(char *)
 *
 *  The buffer is filled in as follows:
 *  h_aliases pointer array - contains pointers into the storage for hostnames.
 *  h_addr_list pointer array - contains pointers into the h_addr_list area.
 *  Filling the buffer this way allows for proper alignment of the h_addr_list
 *  addresses.
 *
 *  This arrangement allows us to alias a Hostent struct pointer as a
 *  real struct hostent* without lying. It also allows us to change the
 *  values contained in the cached entries and requests without changing
 *  the actual hostent pointer, which is saved in a client struct and can't
 *  be changed without blowing things up or a lot more fiddling around.
 *  It also allows for defered allocation of the fixed size buffers until
 *  they are really needed.
 *  Nov. 17, 1997 --Bleep
 */

typedef struct _myres_req			MYRES_REQ;
typedef struct _myres_cache			MYRES_CACHE;
typedef struct _myres_cachetable	MYRES_CACHETABLE;

struct _myres_req
{
	int				id;
	int				sent;		/* number of requests sent */
	int				srch;
	time_t			ttl;
	char			type;
	char			retries;		/* retry counter */
	char			sends;		/* number of sends (>1 means resent) */
	char			resend;		/* send flag. 0 == dont resend */
	time_t			sentat;
	time_t			timeout;
	struct in_addr	addr;
	char			*name;
	MYRES_STATUS	*sptr;
	MYRES_HOSTENT	he;
	MYRES_REQ		*prev;
	MYRES_REQ		*next;
};

struct _myres_cache
{
	time_t			expireat;
	time_t			ttl;
	MYRES_HOSTENT	he;
	MYRES_CACHE		*hname_next;
	MYRES_CACHE		*hnum_next;
	MYRES_CACHE		*next;
};

struct _myres_cachetable
{
	MYRES_CACHE *num_list;
	MYRES_CACHE *name_list;
};

#if 0
static int					_myres_incache = 0;
#endif
static int					_myres_fd = -1;
static MYRES_REQ			*_myres_requests = NULL;
static MYRES_REQ			*_myres_lastrequest = NULL;

static int _myres_proc_answer(MYRES_REQ *, HEADER *, u_char *, u_char *);
static int _myres_query(char *, int, int, MYRES_REQ *);
static int _myres_queryname(MYRES_STATUS *cptr, char *, MYRES_REQ **);
static int _myres_querynumber(MYRES_STATUS *cptr, struct in_addr *, MYRES_REQ **);
static void _myres_resendquery(MYRES_REQ *);

static MYRES_HOSTENT *_myres_gethostbyname(char *name, MYRES_STATUS *cptr, MYRES_REQ **reqptr);
static MYRES_HOSTENT *_myres_gethostbyaddr(struct in_addr *addr, MYRES_STATUS *cptr, MYRES_REQ **reqptr);

static MYRES_REQ *_myres_request_add(MYRES_STATUS *);
static void _myres_request_del(MYRES_REQ *);
static MYRES_REQ *_myres_request_find_id(int);
#if 0
static int _myres_hostent_update(MYRES_HOSTENT *hp, char **addr, char **alias);
#endif

static int myres_sendmsg(char *, int, int);

static struct resinfo
{
	int re_errors;
	int re_nu_look;
	int re_na_look;
	int re_replies;
	int re_requests;
	int re_resends;
	int re_sent;
	int re_timeouts;
	int re_shortttl;
	int re_unkrep;
}
reinfo;


#define IRCD_GETSHORT(__x)	((unsigned short)((unsigned char *)(__x))[0]<<8)| \
							((unsigned short)((unsigned char *)(__x))[1])

#define IRCD_GETLONG(__x)	((unsigned int)((unsigned char *)(__x))[0]<<24)| \
							((unsigned int)((unsigned char *)(__x))[1]<<16)| \
							((unsigned int)((unsigned char *)(__x))[2]<<8)| \
							((unsigned int)((unsigned char *)(__x))[3])

/*
 * Remove a request from the list. This must also free any memory that has
 * been allocated for temporary storage of DNS results.
 */
static void _myres_request_del(MYRES_REQ *old_request)
{
	MYRES_REQ	*keep;

	if (old_request)
	{
		if ((keep = old_request->prev) != NULL)
			old_request->prev->next = old_request->next;
		else
			_myres_requests = old_request->next;
		if (old_request->next)
			old_request->next->prev = old_request->prev;
		else
			_myres_lastrequest = keep;
		if (old_request->he.buf)
			MyFree(old_request->he.buf);
		if (old_request->name)
			MyFree(old_request->name);
		MyFree(old_request);
	}
}

/*
 * Create a DNS request record for the server.
 */
static MYRES_REQ *_myres_request_add(MYRES_STATUS *cptr)
{
	MYRES_REQ *nreq;

	if ((nreq = (MYRES_REQ *) MyMalloc(sizeof(MYRES_REQ))) == NULL)
		return NULL;
	BZERO(nreq, sizeof(MYRES_REQ));
	nreq->sentat = Now;
	nreq->retries = 3;
	nreq->resend = 1;
	nreq->srch = -1;
	nreq->sptr = cptr;
	nreq->timeout = 4;	/* start at 4 and exponential inc. */
	nreq->addr.s_addr = INADDR_NONE;
	nreq->he.h.h_addrtype = AF_INET;
	nreq->he.h.h_length = sizeof(struct in_addr);
	nreq->next = NULL;
	if (!_myres_requests)
	{
		_myres_requests = nreq;
		nreq->prev = NULL;
	}
	else
	{
		_myres_lastrequest->next = nreq;
		nreq->prev = _myres_lastrequest;
	}
	_myres_lastrequest = nreq;
	reinfo.re_requests++;
	return nreq;
}

/*
 * send_res_msg
 *
 * sends msg to all nameservers found in the "_res" structure.
 * This should reflect /etc/resolv.conf. We will get responses
 * which arent needed but is easier than checking to see if nameserver
 * isnt present. Returns number of messages successfully sent to
 * nameservers or -1 if no successful sends.
 */
static int myres_sendmsg(char *msg, int len, int rcount)
{
	int i;
	int sent = 0, max;

	if (!msg)
		return -1;

	max = MYRES_MIN(_res.nscount, rcount);
	if (_res.options & RES_PRIMARY)
		max = 1;
	if (!max)
		max = 1;

	for (i = 0; i < max; ++i)
	{
		_res.nsaddr_list[i].sin_family = AF_INET;
		if (sendto(_myres_fd, msg, len, 0, (struct sockaddr *)
			 &(_res.nsaddr_list[i]), sizeof(struct sockaddr)) == len)
		{
			reinfo.re_sent++;
			sent++;
		}
	}
	return (sent) ? sent : -1;
}

/*
 * find a dns request id (id is determined by dn_mkquery)
 */
static MYRES_REQ *_myres_request_find_id(int id)
{
	MYRES_REQ *rptr;

	for (rptr = _myres_requests;rptr;rptr = rptr->next)
		if (rptr->id == id)
			return rptr;
	return NULL;
}

/*
 * generate a query based on class, type and name.
 */
static int _myres_query(char *name, int q_class, int type, MYRES_REQ *rptr)
{
	struct timeval	tv;
	char			buf[MAXPACKET];
	int				r;
	int				s;
	int				k = 0;
	HEADER			*hptr;

	BZERO(buf, sizeof(buf));
	r = res_mkquery(QUERY, name, q_class, type, NULL, 0, NULL,
		(u_char *) buf, sizeof(buf));
	if (r <= 0)
	{
		h_errno = NO_RECOVERY;
		return r;
	}
	hptr = (HEADER *) buf;
	gettimeofday(&tv, NULL);
	do
	{
		/* htons/ntohs can be assembler macros, which cannot
			 be nested. Thus two lines.	 -Vesa */
		u_short nstmp = ntohs(hptr->id) + k + (u_short) (tv.tv_usec & 0xffff);
		hptr->id = htons(nstmp);
		k++;
	} while (_myres_request_find_id(ntohs(hptr->id)));
	rptr->id = ntohs(hptr->id);
	rptr->sends++;
	s = myres_sendmsg(buf, r, rptr->sends);
	if (s == -1)
	{
		h_errno = TRY_AGAIN;
		return -1;
	}
	else
		rptr->sent += s;
	return 0;
}

static int _myres_queryname(MYRES_STATUS *cptr, char *name, MYRES_REQ **rptr)
{
	char		hname[256];
	int			len = 0;
	char		*sptr = name;
	char		*tptr = hname;
	int			gotdot = 0;
	MYRES_REQ	*myrptr = NULL;

	for(;*sptr && len < (sizeof(hname)-1);len++)
		if ((*tptr++ = *sptr++) == '.')
			gotdot++;
	*tptr = '\0';

	/* only do below if there is room for 2 characters + NULL */
    if ((sizeof(hname)-len) > 2)
	{
		if (rptr && !gotdot && _res.options & RES_DEFNAMES)
		{
			*(hname+len) = '.';
			strncpy(hname+len+1, _res.defdname, sizeof(hname) - len - 2);
			/* NULL is already set above */
		}
	}

	if (!rptr)
		rptr = &myrptr;
	/*
	** Store the name passed as the one to lookup and generate other host
	** names to pass onto the nameserver(s) for lookups.
	*/
	if (!*rptr)
	{
		if ((*rptr = _myres_request_add(cptr)) == NULL)
			return -1;
		cptr->rptr = (void *)(*rptr);
		(*rptr)->type = T_A;
		(*rptr)->name = MyStrdup(name);
	}
	return (_myres_query(hname, C_IN, T_A, *rptr));
}

/*
 * Use this to do reverse IP# lookups.
 */
static int _myres_querynumber(MYRES_STATUS *cptr, struct in_addr *numb, MYRES_REQ **rptr)
{
	char		ipbuf[32];
	u_char		*cp = (u_char *) &numb->s_addr;
	MYRES_REQ	*myrptr = NULL;

	if (!rptr)
		rptr = &myrptr;
	sprintf(ipbuf, "%u.%u.%u.%u.in-addr.arpa.",
		(u_int) (cp[3]), (u_int) (cp[2]),
		(u_int) (cp[1]), (u_int) (cp[0]));
	if (!*rptr)
	{
		if ((*rptr = _myres_request_add(cptr)) == NULL)
			return -1;
		cptr->rptr = (void *)(*rptr);
		(*rptr)->type = T_PTR;
		(*rptr)->addr.s_addr = numb->s_addr;
	}
	return (_myres_query(ipbuf, C_IN, T_PTR, *rptr));
}

static void _myres_resendquery(MYRES_REQ *rptr)
{
	if (rptr->resend == 0)
		return;
	reinfo.re_resends++;
	switch (rptr->type)
	{
		case T_PTR:
			_myres_querynumber(NULL, &rptr->addr, &rptr);
			break;
		case T_A:
			_myres_queryname(NULL, rptr->name, &rptr);
			break;
		default:
			break;
	}
	return;
}

/*
** myres_proc_answer
**
** Process name server reply.
*/
static int _myres_proc_answer(MYRES_REQ *rptr, HEADER * hptr, u_char *buf, u_char *eob)
{
	u_char	*cp = buf + sizeof(HEADER);
	char	**alias;
	char	**addr;
	char	*p;		/* pointer to strings */
	char	*a;		/* pointer to address list */
	char	*endp;		/* end of our buffer */
	int		addr_class;
	int		type;
	int		dlen;
	int		ans = 0;
	int		n;
	int		addr_count = 0;
	int		alias_count = 0;
	int		len;
	struct hostent *hp = (struct hostent *) &(rptr->he.h);
	char	hostbuf[256];

	/*
	 * Lazy allocation of rptr->he.buf, we don't allocate a buffer
	 * unless there's something to put in it.
	 */
	if (!rptr->he.buf)
	{
		if ((rptr->he.buf = MyMalloc(MAXGETHOSTLEN)) == NULL)
			return 0;
		BZERO(rptr->he.buf, MAXGETHOSTLEN);
		/* 
		 * Array of alias list pointers starts at beginning of buf 
		 */
		hp->h_aliases = (char **) rptr->he.buf;
		/* 
		 * Array of address list pointers starts after alias list pointers.
		 * The actual addresses follow the address list pointers.
		 */
		hp->h_addr_list = (char **) (((char *)hp->h_aliases) + ALIASBLEN);
		a = ((char *)hp->h_addr_list) + ADDRSBLEN;
		/* copy the host address to the beginning of h_addr_list */
		if (rptr->addr.s_addr != INADDR_NONE)
		{
			BCOPY(&(rptr->addr), a, sizeof(struct in_addr));
			*hp->h_addr_list = a;
		}
	}
	endp = &rptr->he.buf[MAXGETHOSTLEN];

	/* find the end of the address list */
	addr = hp->h_addr_list;
	while (*addr)
	{
		++addr;
		++addr_count;
	}
	/* make a point to the first available empty address slot */
	a = ((char *)hp->h_addr_list) + ADDRSBLEN +
					(addr_count * sizeof(struct in_addr));
	/* find the end of the alias list */
	alias = hp->h_aliases;
	while (*alias)
	{
		++alias;
		++alias_count;
	}
	/* make p point to the first available space in rptr->buf */
	if (alias_count)
	{
		p = MYRES_MAX(hp->h_name, hp->h_aliases[alias_count - 1]);
		p += (strlen(p) + 1);
	}
	else if (hp->h_name)
		p = hp->h_name + strlen(hp->h_name) + 1;
	else
		p = ((char *)hp->h_addr_list) + ADDRSBLEN + ADDRSDLEN;

	/*
	 * Skip past query's
	 */
	for (; hptr->qdcount > 0; hptr->qdcount--)
	{
		if ((n = dn_skipname(cp, eob)) == -1)
			break;
		else
			cp += (n + QFIXEDSZ);
	}
	/*
	** Proccess each answer sent to us blech.
	*/
	while (hptr->ancount-- > 0 && cp && cp < eob && p < endp)
	{
		if ((n = dn_expand(buf, eob, cp, hostbuf, sizeof(hostbuf))) <= 0)
			break;

		cp += n;
		/* XXX magic numbers, this checks for truncated packets */
		if ((cp + INT16SZ + INT16SZ + INT32SZ + INT16SZ) >= eob)
			break;

		type = IRCD_GETSHORT(cp);
		cp += INT16SZ;
		addr_class = IRCD_GETSHORT(cp);
		cp += INT16SZ;
		rptr->ttl = IRCD_GETLONG(cp);
		cp += INT32SZ;
		dlen = IRCD_GETSHORT(cp);
		cp += INT16SZ;
		rptr->type = type;

		/* check for bad dlen */
		if ((cp + dlen) > eob)
			break;

		/* 
		 * Add default domain name to returned host name if host name
		 * doesn't contain any dot separators.
		 * Name server never returns with trailing '.'
		 */
		if (!strchr(hostbuf, '.') && (_res.options & RES_DEFNAMES))
		{
			len = strlen(hostbuf);
			if ((sizeof(hostbuf)-len) > 2)
			{
				*(hostbuf+len) = '.';
				strncpy(hostbuf+len+1, _res.defdname,
					sizeof(hostbuf) - len - 2);
				hostbuf[sizeof(hostbuf)-1] = 0;
			}
		}

		switch (type)
		{
			case T_A:
				/* check for invalid dlen or too many addresses */
				if ((dlen != sizeof(struct in_addr)) ||
							(++addr_count >= RES_MAXADDRS))
					break;
				if (ans == 1)
					hp->h_addrtype = (addr_class == C_IN) ? AF_INET : AF_UNSPEC;

				BCOPY(cp, a, sizeof(struct in_addr));
				*addr++ = a;
				*addr = 0;
				a += sizeof(struct in_addr);

				if (!hp->h_name)
				{
					len = strlen(hostbuf) + 1;
					if (len > (endp - p))
						break;
					strcpy(p, hostbuf);
					hp->h_name = p;
					p += len+1;
				}
				cp += dlen;
				ans++;
				break;
			case T_PTR:
				if ((n = dn_expand(buf, eob, cp, hostbuf, sizeof(hostbuf))) < 0)
				{
					cp = NULL;
					break;
				}
				cp += n;
			/*
			 * Copy the returned hostname into the host name or alias field if
			 * there is a known hostname already.
			 */
				len = strlen(hostbuf) + 1;
				if (len > (endp - p))
					break;
				if (hp->h_name)
				{
					if (++alias_count >= RES_MAXALIASES)
						break;
					strcpy(p, hostbuf);
					*alias++ = p;
					*alias = NULL;
				}
				else
				{
					strcpy(p, hostbuf);
					hp->h_name = p;
				}
				p += len;
				ans++;
				break;
			case T_CNAME:
				cp += dlen;
				if (++alias_count >= RES_MAXALIASES)
					break;
				len = strlen(hostbuf) + 1;
				if (len > (endp - p))
					break;
				strcpy(p, hostbuf);
				*alias++ = p;
				*alias = NULL;
				p += len;
				ans++;
				break;
			default:
				break;
		}
	}
	if (ans && !hp->h_name)
		hp->h_name = hp->h_aliases[0];/* should have to be true */
	return ans;
}

static MYRES_HOSTENT *_myres_gethostbyname(char *name, MYRES_STATUS *cptr, MYRES_REQ **reqptr)
{
	reinfo.re_na_look++;
	if (cptr)
		_myres_queryname(cptr, name, reqptr);
	return NULL;
}

static MYRES_HOSTENT *_myres_gethostbyaddr(struct in_addr *addr, MYRES_STATUS *cptr, MYRES_REQ **reqptr)
{
	reinfo.re_nu_look++;
	if (cptr)
		_myres_querynumber(cptr, addr, reqptr);
	return NULL;
}

int myres_init(void)
{
	int on = 1;
	int fd = -1;

	BZERO(&reinfo, sizeof(reinfo));

	/* res_init() always returns 0 */
	(void) res_init();

	if (!_res.nscount)
	{
		_res.nscount = 1;
		_res.nsaddr_list[0].sin_addr.s_addr = inet_addr("127.0.0.1");
	}
#ifdef DEBUGMODE
	_res.options |= RES_DEBUG;
#endif

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
		return -errno;
	setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char *) &on, sizeof(on));
	return (_myres_fd = fd);
}

MYRES_HOSTENT *myres_gethostbyname(char *name, MYRES_STATUS *cptr)
{
	MYRES_HOSTENT	*he;
	MYRES_HOSTENT	*henew = NULL;

	he = _myres_gethostbyname(name, cptr, NULL);
	if (he) /* cached...let's copy */
	{
		henew = (MYRES_HOSTENT *)MyMalloc(sizeof(MYRES_HOSTENT));
		if (henew)
			myres_hostent_dup(henew, he);
	}
	return henew;
}

MYRES_HOSTENT *myres_gethostbyaddr(struct in_addr *addr, MYRES_STATUS *cptr)
{
	MYRES_HOSTENT	*he;
	MYRES_HOSTENT	*henew = NULL;

	he = _myres_gethostbyaddr(addr, cptr, NULL);
	if (he) /* cached...let's copy */
	{
		henew = (MYRES_HOSTENT *)MyMalloc(sizeof(MYRES_HOSTENT));
		if (henew)
			myres_hostent_dup(henew, he);
	}
	return henew;
}

void myres_status_init(MYRES_STATUS *cptr, void *ptr, int type)
{
	if (cptr)
	{
		BZERO(cptr, sizeof(MYRES_STATUS));
		cptr->cptr = ptr;
		cptr->ctype = type;
	}
}

void myres_status_deinit(MYRES_STATUS *cptr)
{
	if (cptr)
	{
		if (cptr->rptr)
			_myres_request_del((MYRES_REQ *)cptr->rptr);
		if (cptr->he.buf)
			MyFree(cptr->he.buf);
		cptr->rptr = NULL;
		cptr->he.buf = NULL;
	}
}

/*
 * Read a dns reply from the nameserver and process it.
 */
MYRES_STATUS *myres_get_answer(void)
{
	static u_char		buf[sizeof(HEADER) + MAXPACKET];
	HEADER				*hptr;
	MYRES_REQ			*rptr = NULL;
	struct sockaddr_in	sin;
	int					rc;
	int					a;
	int					max;
	int					len = sizeof(sin);


	for(;;) /* fake loop to avoid goto */
	{
		rc = recvfrom(_myres_fd, buf, sizeof(buf),
						0, (struct sockaddr *) &sin, &len);
		if (rc <= sizeof(HEADER))
			return NULL;
		/*
		** Convert DNS reply reader from Network byte order to CPU byte order.
		*/
		hptr = (HEADER *) buf;
		hptr->id = ntohs(hptr->id);
		hptr->ancount = ntohs(hptr->ancount);
		hptr->qdcount = ntohs(hptr->qdcount);
		hptr->nscount = ntohs(hptr->nscount);
		hptr->arcount = ntohs(hptr->arcount);
		reinfo.re_replies++;
	
		/*
		 * Response for an id which we have already received an answer for
		 * just ignore this response.
		 */
		if ((rptr = _myres_request_find_id(hptr->id)) == NULL)
			return NULL;
		/*
		 * Check against possibly fake replies
		 */
		max = MYRES_MIN(_res.nscount, rptr->sends);
		if (!max)
			max = 1;
	
		for (a = 0; a < max; a++)
		{
			if (!_res.nsaddr_list[a].sin_addr.s_addr ||
					!BCMP((char *) &sin.sin_addr,
					(char *) &_res.nsaddr_list[a].sin_addr, sizeof(struct in_addr)))
				break;
		}
		if (a == max)
		{
			reinfo.re_unkrep++;
			return NULL;
		}
	
		if ((hptr->rcode != NOERROR) || (hptr->ancount == 0))
		{
			switch (hptr->rcode)
			{
				case NXDOMAIN:
					h_errno = TRY_AGAIN;
					break;
				case SERVFAIL:
					h_errno = TRY_AGAIN;
					break;
				case NOERROR:
					h_errno = NO_DATA;
					break;
				case FORMERR:
				case NOTIMP:
				case REFUSED:
				default:
					h_errno = NO_RECOVERY;
					break;
			}
			reinfo.re_errors++;
			/*
			 * If a bad error was returned, we stop here and dont send
			 * send any more (no retries granted).
			 */
			if (h_errno != TRY_AGAIN)
			{
				rptr->resend = 0;
				rptr->retries = 0;
			}
			break;
		}
		/* 
		 * If this fails we didn't get a buffer to hold the hostent or
		 * there was an error decoding the received packet, try it again
		 * and hope it works the next time.
		 */
		a = _myres_proc_answer(rptr, hptr, buf, &buf[rc]);
	
		if (a > 0)
		{
			MYRES_STATUS	*sptr = rptr->sptr;

			sptr->type = rptr->type == T_PTR ? MYRES_TYPE_PTR : MYRES_TYPE_A;
			sptr->result = 1;
			myres_hostent_move(&(sptr->he), &(rptr->he));
			_myres_request_del(rptr);
			return sptr;
		}
		else if (!rptr->sent)
			_myres_request_del(rptr);
		return NULL;
	} /* end of for(;;) */
	/*
	** Reprocess an error if the nameserver didnt tell us to "TRY_AGAIN".
	*/
	if (rptr)
	{
		if (h_errno != TRY_AGAIN)
		{
			/*
			 * If we havent tried with the default domain and its
			 * set, then give it a try next.
			 */
			if (_res.options & RES_DEFNAMES && ++rptr->srch == 0)
			{
				rptr->retries = _res.retry;
				rptr->sends = 0;
				rptr->resend = 1;
				_myres_resendquery(rptr);
			}
			else
				_myres_resendquery(rptr);
		}
		else
		{
			MYRES_STATUS *sptr = rptr->sptr;

			sptr->result = 0;
			_myres_request_del(rptr);
			return sptr;
		}
	}
	return NULL;
}

int myres_hostent_move(MYRES_HOSTENT *new_hp, MYRES_HOSTENT *old_hp)
{
	if (!new_hp || !old_hp)
		return -EINVAL;
	new_hp->buf = old_hp->buf;
	BCOPY(&(old_hp->h), &(new_hp->h), sizeof(struct hostent));
	old_hp->buf = NULL;
	return 0;
}

/*
 * Duplicate a hostent struct, allocate only enough memory for
 * the data we're putting in it.
 */
int myres_hostent_dup(MYRES_HOSTENT *new_hp, MYRES_HOSTENT *old_hp)
{
	char				*p;
	char				**ap;
	char				**pp;
	int					alias_count = 0;
	int					addr_count = 0;
	size_t				bytes_needed = 0;
	struct hostent		*hp;

	if (!new_hp || !old_hp)
		return -EINVAL;
	hp = &(old_hp->h);

	/* how much buffer do we need? */
	bytes_needed += (strlen(hp->h_name) + 1);

	pp = hp->h_aliases;
	while (*pp)
	{
		bytes_needed += (strlen(*pp++) + 1 + sizeof(char *));
		++alias_count;
	}
	pp = hp->h_addr_list;
	while (*pp++)
	{
		bytes_needed += (hp->h_length + sizeof(char *));
		++addr_count;
	}
	/* Reserve space for 2 nulls to terminate h_aliases and h_addr_list */
	bytes_needed += (2 * sizeof(char *));

	/* Allocate memory */
	if ((new_hp->buf = (char *) MyMalloc(bytes_needed)) == NULL)
		return -ENOMEM;

	new_hp->h.h_addrtype = hp->h_addrtype;
	new_hp->h.h_length = hp->h_length;

	/* first write the address list */
	pp = hp->h_addr_list;
	ap = new_hp->h.h_addr_list =
		(char **) (new_hp->buf + ((alias_count + 1) * sizeof(char *)));
	p = (char *) ((char *)ap) + ((addr_count + 1) * sizeof(char *));
	while (*pp)
	{
		*ap++ = p;
		BCOPY(*pp++, p, hp->h_length);
		p += hp->h_length;
	}
	*ap = NULL;
	/* next write the name */
	new_hp->h.h_name = p;
	strcpy(p, hp->h_name);
	p += (strlen(p) + 1);
	/* last write the alias list */
	pp = hp->h_aliases;
	ap = new_hp->h.h_aliases = (char **) new_hp->buf;
	while (*pp)
	{
		*ap++ = p;
		strcpy(p, *pp++);
		p += (strlen(p) + 1);
	}
	*ap = NULL;
	return 0;
}

int myres_hostent_free(MYRES_HOSTENT *heptr)
{
	if (heptr && heptr->buf)
		MyFree(heptr->buf);
	return 0;
}

int myres_retry_request(void *reqptr)
{
	MYRES_REQ	*rptr = (MYRES_REQ *)reqptr;

	if (!rptr)
		return 1;

	if (Now >= (rptr->sentat + rptr->timeout))
	{
		if (--rptr->retries <= 0)
			return 1;
		rptr->sentat = Now;
		rptr->timeout += rptr->timeout;
		_myres_resendquery(rptr);
	}
	return 0;
}

